global string s1, s2;
global int tt;

public class Receipt extends GGen
{
    string reg1, reg2;
    int type; // should be 1 for plus, -1 for minus
    GText t --> this; 

    Mouse @ mouse;    
    MusicBox @ musicbox;
    
    GPlane pad --> this;
    FlatMaterial mat;
    pad.mat(mat);
    pad.sca( @(4.8, 0.8, 0) );

    // events
    Event onHoverEvent, onClickEvent, onExitEvent, onReleaseEvent;  // onExit, onRelease
    
    // states
    0 => static int NONE;  // not hovered or active
    1 => static int HOVERED;  // hovered
    2 => static int ACTIVE;   // clicked
    // 3 => static int PLAYING;  // makine sound!
    0 => int state; // current state

    // input types
    0 => static int MOUSE_HOVER;
    1 => static int MOUSE_EXIT;
    2 => static int MOUSE_CLICK;
    3 => static int MOUSE_RELEASE;
    // 3 => static int NOTE_ON;
    // 4 => static int NOTE_OFF;

    [
        @(237, 233, 216)/255,    // NONE
        Color.WHITE,      // HOVERED
        Color.WHITE, // ACTIVE
        Color.GREEN     // PLAYING
    ] @=> vec3 colorMap[];


    // constructor
    fun void init(Mouse @ m, MusicBox @ mb, string r1, int type, string r2, string result)
    {

        if (mouse != null) return;
        m @=> this.mouse;

        mb @=> this.musicbox;

        // e @=> this.onReleaseEvent;

        spork ~ this.clickListener();
        spork ~ this.releaseListener();

        r1 => reg1;
        r2 => reg2;
        type => this.type;
        
        string operation;
        if(type > 0)
        {
            "+" => operation;
        }
        else
        {
            "-" => operation;
        }

        t.text(r1 + " " + operation + " " + r2 + " = " + result );

        <<<"here">>>;
        t.color( @(0, 0, 0) );
        t.sca( 0.3 );
        t.pos( @(0, 0, 1) );
   
    }
    
    // check if state is active (i.e. should play sound)
    fun int isActive() {
        return state == ACTIVE;
    }

    // set color
    fun void setColor(vec3 c) {
        pad.mat().color(c);
    }

    // returns true if mouse is hovering over pad
    fun int isHovered() {
        pad.scaWorld() => vec3 worldScale;  // get dimensions
        worldScale.x / 2.0 => float halfWidth;
        worldScale.y / 2.0 => float halfHeight;
        pad.posWorld() => vec3 worldPos;   // get position

        // <<<worldScale>>>;
        // <<<mouse.worldPos>>>;
        if (mouse.worldPos.x > worldPos.x - halfWidth && mouse.worldPos.x < worldPos.x + halfWidth &&
            mouse.worldPos.y > worldPos.y - halfHeight && mouse.worldPos.y < worldPos.y + halfHeight) {
            return true;
        }
        return false;
    }

    // poll for hover events
    fun void checkHover() {
        // <<<isHovered()>>>;
        if (isHovered()) {
            handleInput(MOUSE_HOVER);
        } 
        else {
            if (state == HOVERED) handleInput(MOUSE_EXIT);
        }
    }

    // handle mouse clicks
    fun void clickListener() {
        now => time lastClick;
        while (true) {
            mouse.mouseDownEvents[Mouse.LEFT] => now;
            if (isHovered()) {
                handleInput(MOUSE_CLICK);
            }
            100::ms => now; // cooldown
        }
    }

    fun void releaseListener() {
        now => time lastRelease;
        while (true) {
            mouse.mouseUpEvents[Mouse.LEFT] => now;
            if (isActive()) {
                // onReleaseEvent.broadcast();
                handleInput(MOUSE_RELEASE);
                // myDigit => currInput;
            }
            100::ms => now; // cooldown
        }
    }

    // animetion when clicked
    fun void animate(){
        if (state == ACTIVE)
        {
            this.sca(0.9);
        }
        else
        {
            this.sca(1);
        }
    }



    // activate pad, meaning it should be played when the sequencer hits it
    fun void activate() {
        enter(ACTIVE);
    }

    0 => int lastState;
    // enter state, remember last state
    fun void enter(int s) {
        state => lastState;
        s => state;
    }

    // basic state machine for handling input
    MusicBox @ lastmb;

    fun void handleInput(int input) {

        // <<<input>>>;
        if (state == NONE) {
            if (input == MOUSE_HOVER)      enter(HOVERED);
            else if (input == MOUSE_CLICK)
            { 
                enter(ACTIVE);
            }
        } 
        
        else if (state == HOVERED) {
            if (input == MOUSE_EXIT)       enter(NONE);
            else if (input == MOUSE_CLICK)
            { 
                enter(ACTIVE);
                
                // do something here
                reg1 => s1;
                reg2 => s2;
                type => tt;
                // lastmb.stopPlay();
                
                // musicbox.setScore(reg1, reg2, type);
                // musicbox @=> lastmb;

                <<<"Already playing:" + musicbox.isPlaying()>>>;
                musicbox.setScore();
                if(musicbox.isPlaying() != 1)
                {
                    spork ~ musicbox.startPlay();
                }
                
            }
        } 
        
        else if (state == ACTIVE) {
            // if (input == MOUSE_CLICK)      enter(NONE);
            if( input == MOUSE_RELEASE ){
                enter(NONE);
                onReleaseEvent.broadcast();
            }
        } 
    }

    // override ggen update
    fun void update(float dt) {
        // <<<"here">>>;
        // check if hovered
        checkHover();
        animate();
        // update state
        // <<<"state", state>>>;
        this.setColor(colorMap[state]);
    }
}